package xyz.lychee.lagfixer.libs.adventure.text;

import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.Spliterator;
import java.util.Spliterators;
import java.util.function.BiPredicate;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.UnaryOperator;
import java.util.regex.Pattern;
import java.util.stream.Collector;
import java.util.stream.Stream;
import xyz.lychee.lagfixer.libs.adventure.builder.AbstractBuilder;
import xyz.lychee.lagfixer.libs.adventure.key.Key;
import xyz.lychee.lagfixer.libs.adventure.text.event.ClickEvent;
import xyz.lychee.lagfixer.libs.adventure.text.event.HoverEvent;
import xyz.lychee.lagfixer.libs.adventure.text.event.HoverEventSource;
import xyz.lychee.lagfixer.libs.adventure.text.format.Style;
import xyz.lychee.lagfixer.libs.adventure.text.format.StyleBuilderApplicable;
import xyz.lychee.lagfixer.libs.adventure.text.format.StyleGetter;
import xyz.lychee.lagfixer.libs.adventure.text.format.StyleSetter;
import xyz.lychee.lagfixer.libs.adventure.text.format.TextColor;
import xyz.lychee.lagfixer.libs.adventure.text.format.TextDecoration;
import xyz.lychee.lagfixer.libs.adventure.text.format.Style.Merge;
import xyz.lychee.lagfixer.libs.adventure.text.format.Style.Merge.Strategy;
import xyz.lychee.lagfixer.libs.adventure.text.format.TextDecoration.State;
import xyz.lychee.lagfixer.libs.adventure.translation.Translatable;
import xyz.lychee.lagfixer.libs.adventure.util.ForwardingIterator;
import xyz.lychee.lagfixer.libs.adventure.util.IntFunction2;
import xyz.lychee.lagfixer.libs.adventure.util.MonkeyBars;
import xyz.lychee.lagfixer.libs.annotations.Contract;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;
import xyz.lychee.lagfixer.libs.annotations.Unmodifiable;
import xyz.lychee.lagfixer.libs.annotations.ApiStatus.NonExtendable;
import xyz.lychee.lagfixer.libs.annotations.ApiStatus.ScheduledForRemoval;
import xyz.lychee.lagfixer.libs.examination.Examinable;
import xyz.lychee.lagfixer.libs.examination.ExaminableProperty;

@NonExtendable
public interface Component extends ComponentBuilderApplicable, ComponentLike, Examinable, HoverEventSource<Component>, StyleGetter, StyleSetter<Component> {
   BiPredicate<? super Component, ? super Component> EQUALS = Objects::equals;
   BiPredicate<? super Component, ? super Component> EQUALS_IDENTITY = (a, b) -> {
      return a == b;
   };
   Predicate<? super Component> IS_NOT_EMPTY = (component) -> {
      return component != empty();
   };

   static @NotNull TextComponent empty() {
      return TextComponentImpl.EMPTY;
   }

   static @NotNull TextComponent newline() {
      return TextComponentImpl.NEWLINE;
   }

   static @NotNull TextComponent space() {
      return TextComponentImpl.SPACE;
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent join(final @NotNull ComponentLike separator, final @NotNull ComponentLike... components) {
      return join((ComponentLike)separator, (Iterable)Arrays.asList(components));
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent join(final @NotNull ComponentLike separator, final Iterable<? extends ComponentLike> components) {
      Component component = join(JoinConfiguration.separator(separator), components);
      return component instanceof TextComponent ? (TextComponent)component : (TextComponent)((TextComponent.Builder)text().append(component)).build();
   }

   @Contract(
      pure = true
   )
   static @NotNull Component join(final JoinConfiguration.@NotNull Builder configBuilder, final @NotNull ComponentLike... components) {
      return join((JoinConfiguration.Builder)configBuilder, (Iterable)Arrays.asList(components));
   }

   @Contract(
      pure = true
   )
   static @NotNull Component join(final JoinConfiguration.@NotNull Builder configBuilder, final @NotNull Iterable<? extends ComponentLike> components) {
      return JoinConfigurationImpl.join((JoinConfiguration)configBuilder.build(), components);
   }

   @Contract(
      pure = true
   )
   static @NotNull Component join(final @NotNull JoinConfiguration config, final @NotNull ComponentLike... components) {
      return join((JoinConfiguration)config, (Iterable)Arrays.asList(components));
   }

   @Contract(
      pure = true
   )
   static @NotNull Component join(final @NotNull JoinConfiguration config, final @NotNull Iterable<? extends ComponentLike> components) {
      return JoinConfigurationImpl.join(config, components);
   }

   static @NotNull Collector<Component, ? extends ComponentBuilder<?, ?>, Component> toComponent() {
      return toComponent(empty());
   }

   static @NotNull Collector<Component, ? extends ComponentBuilder<?, ?>, Component> toComponent(final @NotNull Component separator) {
      return Collector.of(Component::text, (builder, add) -> {
         if (separator != empty() && !builder.children().isEmpty()) {
            builder.append(separator);
         }

         builder.append(add);
      }, (a, b) -> {
         List<Component> aChildren = a.children();
         TextComponent.Builder ret = (TextComponent.Builder)text().append(aChildren);
         if (!aChildren.isEmpty()) {
            ret.append(separator);
         }

         ret.append(b.children());
         return ret;
      }, ComponentBuilder::build);
   }

   @Contract(
      pure = true
   )
   static BlockNBTComponent.@NotNull Builder blockNBT() {
      return new BlockNBTComponentImpl.BuilderImpl();
   }

   @Contract("_ -> new")
   static @NotNull BlockNBTComponent blockNBT(final @NotNull Consumer<? super BlockNBTComponent.Builder> consumer) {
      return (BlockNBTComponent)AbstractBuilder.configureAndBuild(blockNBT(), consumer);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull BlockNBTComponent blockNBT(final @NotNull String nbtPath, final BlockNBTComponent.@NotNull Pos pos) {
      return blockNBT(nbtPath, false, pos);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull BlockNBTComponent blockNBT(final @NotNull String nbtPath, final boolean interpret, final BlockNBTComponent.@NotNull Pos pos) {
      return blockNBT(nbtPath, interpret, (ComponentLike)null, pos);
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull BlockNBTComponent blockNBT(final @NotNull String nbtPath, final boolean interpret, final @Nullable ComponentLike separator, final BlockNBTComponent.@NotNull Pos pos) {
      return BlockNBTComponentImpl.create(Collections.emptyList(), Style.empty(), nbtPath, interpret, separator, pos);
   }

   @Contract(
      pure = true
   )
   static EntityNBTComponent.@NotNull Builder entityNBT() {
      return new EntityNBTComponentImpl.BuilderImpl();
   }

   @Contract("_ -> new")
   static @NotNull EntityNBTComponent entityNBT(final @NotNull Consumer<? super EntityNBTComponent.Builder> consumer) {
      return (EntityNBTComponent)AbstractBuilder.configureAndBuild(entityNBT(), consumer);
   }

   @Contract("_, _ -> new")
   static @NotNull EntityNBTComponent entityNBT(final @NotNull String nbtPath, final @NotNull String selector) {
      return (EntityNBTComponent)((EntityNBTComponent.Builder)entityNBT().nbtPath(nbtPath)).selector(selector).build();
   }

   @Contract(
      pure = true
   )
   static KeybindComponent.@NotNull Builder keybind() {
      return new KeybindComponentImpl.BuilderImpl();
   }

   @Contract("_ -> new")
   static @NotNull KeybindComponent keybind(final @NotNull Consumer<? super KeybindComponent.Builder> consumer) {
      return (KeybindComponent)AbstractBuilder.configureAndBuild(keybind(), consumer);
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull KeybindComponent keybind(final @NotNull String keybind) {
      return keybind(keybind, Style.empty());
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull KeybindComponent keybind(final KeybindComponent.@NotNull KeybindLike keybind) {
      return keybind(((KeybindComponent.KeybindLike)Objects.requireNonNull(keybind, "keybind")).asKeybind(), Style.empty());
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull KeybindComponent keybind(final @NotNull String keybind, final @NotNull Style style) {
      return KeybindComponentImpl.create(Collections.emptyList(), (Style)Objects.requireNonNull(style, "style"), keybind);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull KeybindComponent keybind(final KeybindComponent.@NotNull KeybindLike keybind, final @NotNull Style style) {
      return KeybindComponentImpl.create(Collections.emptyList(), (Style)Objects.requireNonNull(style, "style"), ((KeybindComponent.KeybindLike)Objects.requireNonNull(keybind, "keybind")).asKeybind());
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull KeybindComponent keybind(final @NotNull String keybind, final @Nullable TextColor color) {
      return keybind(keybind, Style.style(color));
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull KeybindComponent keybind(final KeybindComponent.@NotNull KeybindLike keybind, final @Nullable TextColor color) {
      return keybind(((KeybindComponent.KeybindLike)Objects.requireNonNull(keybind, "keybind")).asKeybind(), Style.style(color));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull KeybindComponent keybind(final @NotNull String keybind, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return keybind(keybind, Style.style(color, decorations));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull KeybindComponent keybind(final KeybindComponent.@NotNull KeybindLike keybind, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return keybind(((KeybindComponent.KeybindLike)Objects.requireNonNull(keybind, "keybind")).asKeybind(), Style.style(color, decorations));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull KeybindComponent keybind(final @NotNull String keybind, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return keybind(keybind, Style.style(color, decorations));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull KeybindComponent keybind(final KeybindComponent.@NotNull KeybindLike keybind, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return keybind(((KeybindComponent.KeybindLike)Objects.requireNonNull(keybind, "keybind")).asKeybind(), Style.style(color, decorations));
   }

   @Contract(
      pure = true
   )
   static ScoreComponent.@NotNull Builder score() {
      return new ScoreComponentImpl.BuilderImpl();
   }

   @Contract("_ -> new")
   static @NotNull ScoreComponent score(final @NotNull Consumer<? super ScoreComponent.Builder> consumer) {
      return (ScoreComponent)AbstractBuilder.configureAndBuild(score(), consumer);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull ScoreComponent score(final @NotNull String name, final @NotNull String objective) {
      return score(name, objective, (String)null);
   }

   /** @deprecated */
   @Deprecated
   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull ScoreComponent score(final @NotNull String name, final @NotNull String objective, final @Nullable String value) {
      return ScoreComponentImpl.create(Collections.emptyList(), Style.empty(), name, objective, value);
   }

   @Contract(
      pure = true
   )
   static SelectorComponent.@NotNull Builder selector() {
      return new SelectorComponentImpl.BuilderImpl();
   }

   @Contract("_ -> new")
   static @NotNull SelectorComponent selector(final @NotNull Consumer<? super SelectorComponent.Builder> consumer) {
      return (SelectorComponent)AbstractBuilder.configureAndBuild(selector(), consumer);
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull SelectorComponent selector(final @NotNull String pattern) {
      return selector(pattern, (ComponentLike)null);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull SelectorComponent selector(final @NotNull String pattern, final @Nullable ComponentLike separator) {
      return SelectorComponentImpl.create(Collections.emptyList(), Style.empty(), pattern, separator);
   }

   @Contract(
      pure = true
   )
   static StorageNBTComponent.@NotNull Builder storageNBT() {
      return new StorageNBTComponentImpl.BuilderImpl();
   }

   @Contract("_ -> new")
   static @NotNull StorageNBTComponent storageNBT(final @NotNull Consumer<? super StorageNBTComponent.Builder> consumer) {
      return (StorageNBTComponent)AbstractBuilder.configureAndBuild(storageNBT(), consumer);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull StorageNBTComponent storageNBT(final @NotNull String nbtPath, final @NotNull Key storage) {
      return storageNBT(nbtPath, false, storage);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull StorageNBTComponent storageNBT(final @NotNull String nbtPath, final boolean interpret, final @NotNull Key storage) {
      return storageNBT(nbtPath, interpret, (ComponentLike)null, storage);
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull StorageNBTComponent storageNBT(final @NotNull String nbtPath, final boolean interpret, final @Nullable ComponentLike separator, final @NotNull Key storage) {
      return StorageNBTComponentImpl.create(Collections.emptyList(), Style.empty(), nbtPath, interpret, separator, storage);
   }

   @Contract(
      pure = true
   )
   static TextComponent.@NotNull Builder text() {
      return new TextComponentImpl.BuilderImpl();
   }

   static @NotNull TextComponent textOfChildren(final @NotNull ComponentLike... components) {
      return components.length == 0 ? empty() : TextComponentImpl.create(Arrays.asList(components), Style.empty(), "");
   }

   @Contract("_ -> new")
   static @NotNull TextComponent text(final @NotNull Consumer<? super TextComponent.Builder> consumer) {
      return (TextComponent)AbstractBuilder.configureAndBuild(text(), consumer);
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final @NotNull String content) {
      return content.isEmpty() ? empty() : text(content, Style.empty());
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final @NotNull String content, final @NotNull Style style) {
      return TextComponentImpl.create(Collections.emptyList(), (Style)Objects.requireNonNull(style, "style"), content);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final @NotNull String content, final @Nullable TextColor color) {
      return text(content, Style.style(color));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final @NotNull String content, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return text(content, Style.style(color, decorations));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final @NotNull String content, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return text(content, Style.style(color, decorations));
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final boolean value) {
      return text(String.valueOf(value));
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final boolean value, final @NotNull Style style) {
      return text(String.valueOf(value), style);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final boolean value, final @Nullable TextColor color) {
      return text(String.valueOf(value), color);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final boolean value, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final boolean value, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      pure = true
   )
   static @NotNull TextComponent text(final char value) {
      if (value == '\n') {
         return newline();
      } else {
         return value == ' ' ? space() : text(String.valueOf(value));
      }
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final char value, final @NotNull Style style) {
      return text(String.valueOf(value), style);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final char value, final @Nullable TextColor color) {
      return text(String.valueOf(value), color);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final char value, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final char value, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final double value) {
      return text(String.valueOf(value));
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final double value, final @NotNull Style style) {
      return text(String.valueOf(value), style);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final double value, final @Nullable TextColor color) {
      return text(String.valueOf(value), color);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final double value, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final double value, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final float value) {
      return text(String.valueOf(value));
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final float value, final @NotNull Style style) {
      return text(String.valueOf(value), style);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final float value, final @Nullable TextColor color) {
      return text(String.valueOf(value), color);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final float value, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final float value, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final int value) {
      return text(String.valueOf(value));
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final int value, final @NotNull Style style) {
      return text(String.valueOf(value), style);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final int value, final @Nullable TextColor color) {
      return text(String.valueOf(value), color);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final int value, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final int value, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final long value) {
      return text(String.valueOf(value));
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final long value, final @NotNull Style style) {
      return text(String.valueOf(value), style);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final long value, final @Nullable TextColor color) {
      return text(String.valueOf(value), color);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final long value, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TextComponent text(final long value, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return text(String.valueOf(value), color, decorations);
   }

   @Contract(
      pure = true
   )
   static TranslatableComponent.@NotNull Builder translatable() {
      return new TranslatableComponentImpl.BuilderImpl();
   }

   @Contract("_ -> new")
   static @NotNull TranslatableComponent translatable(final @NotNull Consumer<? super TranslatableComponent.Builder> consumer) {
      return (TranslatableComponent)AbstractBuilder.configureAndBuild(translatable(), consumer);
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key) {
      return translatable(key, Style.empty());
   }

   @Contract(
      value = "_ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), Style.empty());
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable String fallback) {
      return translatable(key, fallback, Style.empty());
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable String fallback) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), fallback, Style.empty());
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @NotNull Style style) {
      return TranslatableComponentImpl.create(Collections.emptyList(), (Style)Objects.requireNonNull(style, "style"), key, (String)null, (List)Collections.emptyList());
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @NotNull Style style) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), style);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable String fallback, final @NotNull Style style) {
      return TranslatableComponentImpl.create(Collections.emptyList(), (Style)Objects.requireNonNull(style, "style"), key, fallback, Collections.emptyList());
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable String fallback, final @NotNull Style style) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), fallback, style);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable String fallback, final StyleBuilderApplicable... style) {
      return translatable((String)Objects.requireNonNull(key, "key"), fallback, Style.style(style));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable String fallback, final @NotNull Iterable<StyleBuilderApplicable> style) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), fallback, Style.style(style));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable String fallback, final @NotNull ComponentLike... args) {
      return translatable(key, fallback, Style.empty(), args);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable String fallback, final @NotNull ComponentLike... args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), fallback, args);
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable String fallback, final @NotNull Style style, final @NotNull ComponentLike... args) {
      return TranslatableComponentImpl.create(Collections.emptyList(), (Style)Objects.requireNonNull(style, "style"), key, fallback, (ComponentLike[])Objects.requireNonNull(args, "args"));
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable String fallback, final @NotNull Style style, final @NotNull ComponentLike... args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), fallback, style, args);
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable String fallback, final @NotNull Style style, final @NotNull List<? extends ComponentLike> args) {
      return TranslatableComponentImpl.create(Collections.emptyList(), style, key, fallback, (List)Objects.requireNonNull(args, "args"));
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable String fallback, final @NotNull Style style, final @NotNull List<? extends ComponentLike> args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), fallback, style, args);
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable String fallback, final @NotNull List<? extends ComponentLike> args, final @NotNull Iterable<StyleBuilderApplicable> style) {
      return TranslatableComponentImpl.create(Collections.emptyList(), Style.style(style), key, fallback, (List)Objects.requireNonNull(args, "args"));
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable String fallback, final @NotNull List<? extends ComponentLike> args, final @NotNull Iterable<StyleBuilderApplicable> style) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), fallback, args, style);
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable String fallback, final @NotNull List<? extends ComponentLike> args, final StyleBuilderApplicable... style) {
      return TranslatableComponentImpl.create(Collections.emptyList(), Style.style(style), key, fallback, (List)Objects.requireNonNull(args, "args"));
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable String fallback, final @NotNull List<? extends ComponentLike> args, final StyleBuilderApplicable... style) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), fallback, args, style);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable TextColor color) {
      return translatable(key, Style.style(color));
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable TextColor color) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), color);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return translatable(key, Style.style(color, decorations));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable TextColor color, final @NotNull TextDecoration... decorations) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), color, decorations);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return translatable(key, Style.style(color, decorations));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), color, decorations);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @NotNull ComponentLike... args) {
      return translatable(key, Style.empty(), args);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @NotNull ComponentLike... args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), args);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @NotNull Style style, final @NotNull ComponentLike... args) {
      return TranslatableComponentImpl.create(Collections.emptyList(), (Style)Objects.requireNonNull(style, "style"), key, (String)null, (ComponentLike[])((ComponentLike[])Objects.requireNonNull(args, "args")));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @NotNull Style style, final @NotNull ComponentLike... args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), style, args);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable TextColor color, final @NotNull ComponentLike... args) {
      return translatable(key, Style.style(color), args);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable TextColor color, final @NotNull ComponentLike... args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), color, args);
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations, final @NotNull ComponentLike... args) {
      return translatable(key, Style.style(color, decorations), args);
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations, final @NotNull ComponentLike... args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), color, decorations, args);
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @NotNull List<? extends ComponentLike> args) {
      return TranslatableComponentImpl.create(Collections.emptyList(), Style.empty(), key, (String)null, (List)((List)Objects.requireNonNull(args, "args")));
   }

   @Contract(
      value = "_, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @NotNull List<? extends ComponentLike> args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), args);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @NotNull Style style, final @NotNull List<? extends ComponentLike> args) {
      return TranslatableComponentImpl.create(Collections.emptyList(), (Style)Objects.requireNonNull(style, "style"), key, (String)null, (List)((List)Objects.requireNonNull(args, "args")));
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @NotNull Style style, final @NotNull List<? extends ComponentLike> args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), style, args);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static TranslatableComponent translatable(final @NotNull String key, final @Nullable TextColor color, final @NotNull List<? extends ComponentLike> args) {
      return translatable(key, Style.style(color), args);
   }

   @Contract(
      value = "_, _, _ -> new",
      pure = true
   )
   static TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable TextColor color, final @NotNull List<? extends ComponentLike> args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), color, args);
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull String key, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations, final @NotNull List<? extends ComponentLike> args) {
      return translatable(key, Style.style(color, decorations), args);
   }

   @Contract(
      value = "_, _, _, _ -> new",
      pure = true
   )
   static @NotNull TranslatableComponent translatable(final @NotNull Translatable translatable, final @Nullable TextColor color, final @NotNull Set<TextDecoration> decorations, final @NotNull List<? extends ComponentLike> args) {
      return translatable(((Translatable)Objects.requireNonNull(translatable, "translatable")).translationKey(), color, decorations, args);
   }

   @Unmodifiable @NotNull List<Component> children();

   @Contract(
      pure = true
   )
   @NotNull Component children(final @NotNull List<? extends ComponentLike> children);

   default boolean contains(final @NotNull Component that) {
      return this.contains(that, EQUALS_IDENTITY);
   }

   default boolean contains(@NotNull final Component that, @NotNull final BiPredicate<? super Component, ? super Component> equals) {
      if (equals.test(this, that)) {
         return true;
      } else {
         Iterator var3 = this.children().iterator();

         while(var3.hasNext()) {
            Component child = (Component)var3.next();
            if (child.contains(that, equals)) {
               return true;
            }
         }

         HoverEvent<?> hoverEvent = this.hoverEvent();
         if (hoverEvent != null) {
            Object value = hoverEvent.value();
            Component component = null;
            if (value instanceof Component) {
               component = (Component)hoverEvent.value();
            } else if (value instanceof HoverEvent.ShowEntity) {
               component = ((HoverEvent.ShowEntity)value).name();
            }

            if (component != null) {
               if (equals.test(that, component)) {
                  return true;
               }

               Iterator var6 = component.children().iterator();

               while(var6.hasNext()) {
                  Component child = (Component)var6.next();
                  if (child.contains(that, equals)) {
                     return true;
                  }
               }
            }
         }

         return false;
      }
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   default void detectCycle(final @NotNull Component that) {
      if (that.contains(this)) {
         throw new IllegalStateException("Component cycle detected between " + this + " and " + that);
      }
   }

   @Contract(
      pure = true
   )
   default @NotNull Component append(final @NotNull Component component) {
      return this.append((ComponentLike)component);
   }

   default @NotNull Component append(final @NotNull ComponentLike like) {
      Objects.requireNonNull(like, "like");
      Component component = like.asComponent();
      Objects.requireNonNull(component, "component");
      if (component == empty()) {
         return this;
      } else {
         List<Component> oldChildren = this.children();
         return this.children(MonkeyBars.addOne(oldChildren, component));
      }
   }

   @Contract(
      pure = true
   )
   default @NotNull Component append(final @NotNull ComponentBuilder<?, ?> builder) {
      return this.append((Component)builder.build());
   }

   @Contract(
      pure = true
   )
   default @NotNull Component appendNewline() {
      return this.append((Component)newline());
   }

   @Contract(
      pure = true
   )
   default @NotNull Component appendSpace() {
      return this.append((Component)space());
   }

   @Contract(
      pure = true
   )
   default @NotNull Component applyFallbackStyle(final @NotNull Style style) {
      Objects.requireNonNull(style, "style");
      return this.style(this.style().merge(style, Strategy.IF_ABSENT_ON_TARGET));
   }

   @Contract(
      pure = true
   )
   default @NotNull Component applyFallbackStyle(final @NotNull StyleBuilderApplicable... style) {
      return this.applyFallbackStyle(Style.style(style));
   }

   @NotNull Style style();

   @Contract(
      pure = true
   )
   @NotNull Component style(final @NotNull Style style);

   @Contract(
      pure = true
   )
   default @NotNull Component style(final @NotNull Consumer<Style.Builder> consumer) {
      return this.style(this.style().edit(consumer));
   }

   @Contract(
      pure = true
   )
   default @NotNull Component style(final @NotNull Consumer<Style.Builder> consumer, final Style.Merge.@NotNull Strategy strategy) {
      return this.style(this.style().edit(consumer, strategy));
   }

   @Contract(
      pure = true
   )
   default @NotNull Component style(final Style.@NotNull Builder style) {
      return this.style(style.build());
   }

   @Contract(
      pure = true
   )
   default @NotNull Component mergeStyle(final @NotNull Component that) {
      return this.mergeStyle(that, Merge.all());
   }

   @Contract(
      pure = true
   )
   default @NotNull Component mergeStyle(final @NotNull Component that, final Style.@NotNull Merge... merges) {
      return this.mergeStyle(that, Merge.merges(merges));
   }

   @Contract(
      pure = true
   )
   default @NotNull Component mergeStyle(final @NotNull Component that, final @NotNull Set<Style.Merge> merges) {
      return this.style(this.style().merge(that.style(), merges));
   }

   default @Nullable Key font() {
      return this.style().font();
   }

   default @NotNull Component font(final @Nullable Key key) {
      return this.style(this.style().font(key));
   }

   default @Nullable TextColor color() {
      return this.style().color();
   }

   @Contract(
      pure = true
   )
   default @NotNull Component color(final @Nullable TextColor color) {
      return this.style(this.style().color(color));
   }

   @Contract(
      pure = true
   )
   default @NotNull Component colorIfAbsent(final @Nullable TextColor color) {
      return this.color() == null ? this.color(color) : this;
   }

   default boolean hasDecoration(final @NotNull TextDecoration decoration) {
      return super.hasDecoration(decoration);
   }

   @Contract(
      pure = true
   )
   default @NotNull Component decorate(final @NotNull TextDecoration decoration) {
      return (Component)super.decorate(decoration);
   }

   default TextDecoration.@NotNull State decoration(final @NotNull TextDecoration decoration) {
      return this.style().decoration(decoration);
   }

   @Contract(
      pure = true
   )
   default @NotNull Component decoration(final @NotNull TextDecoration decoration, final boolean flag) {
      return (Component)super.decoration(decoration, flag);
   }

   @Contract(
      pure = true
   )
   default @NotNull Component decoration(final @NotNull TextDecoration decoration, final TextDecoration.@NotNull State state) {
      return this.style(this.style().decoration(decoration, state));
   }

   @NotNull
   default Component decorationIfAbsent(@NotNull final TextDecoration decoration, final TextDecoration.State state) {
      Objects.requireNonNull(state, "state");
      TextDecoration.State oldState = this.decoration(decoration);
      return oldState == State.NOT_SET ? this.style(this.style().decoration(decoration, state)) : this;
   }

   default @NotNull Map<TextDecoration, TextDecoration.State> decorations() {
      return this.style().decorations();
   }

   @Contract(
      pure = true
   )
   default @NotNull Component decorations(final @NotNull Map<TextDecoration, TextDecoration.State> decorations) {
      return this.style(this.style().decorations(decorations));
   }

   default @Nullable ClickEvent clickEvent() {
      return this.style().clickEvent();
   }

   @Contract(
      pure = true
   )
   default @NotNull Component clickEvent(final @Nullable ClickEvent event) {
      return this.style(this.style().clickEvent(event));
   }

   default @Nullable HoverEvent<?> hoverEvent() {
      return this.style().hoverEvent();
   }

   @Contract(
      pure = true
   )
   default @NotNull Component hoverEvent(final @Nullable HoverEventSource<?> source) {
      return this.style(this.style().hoverEvent(source));
   }

   default @Nullable String insertion() {
      return this.style().insertion();
   }

   @Contract(
      pure = true
   )
   default @NotNull Component insertion(final @Nullable String insertion) {
      return this.style(this.style().insertion(insertion));
   }

   default boolean hasStyling() {
      return !this.style().isEmpty();
   }

   @Contract(
      pure = true
   )
   default @NotNull Component replaceText(final @NotNull Consumer<TextReplacementConfig.Builder> configurer) {
      Objects.requireNonNull(configurer, "configurer");
      return this.replaceText((TextReplacementConfig)AbstractBuilder.configureAndBuild(TextReplacementConfig.builder(), configurer));
   }

   @Contract(
      pure = true
   )
   default @NotNull Component replaceText(final @NotNull TextReplacementConfig config) {
      Objects.requireNonNull(config, "replacement");
      if (!(config instanceof TextReplacementConfigImpl)) {
         throw new IllegalArgumentException("Provided replacement was a custom TextReplacementConfig implementation, which is not supported.");
      } else {
         return TextReplacementRenderer.INSTANCE.render(this, ((TextReplacementConfigImpl)config).createState());
      }
   }

   default @NotNull Component compact() {
      return ComponentCompaction.compact(this, (Style)null);
   }

   default @NotNull Iterable<Component> iterable(final @NotNull ComponentIteratorType type, final @Nullable ComponentIteratorFlag... flags) {
      return this.iterable(type, flags == null ? Collections.emptySet() : MonkeyBars.enumSet(ComponentIteratorFlag.class, flags));
   }

   default @NotNull Iterable<Component> iterable(final @NotNull ComponentIteratorType type, final @NotNull Set<ComponentIteratorFlag> flags) {
      Objects.requireNonNull(type, "type");
      Objects.requireNonNull(flags, "flags");
      return new ForwardingIterator(() -> {
         return this.iterator(type, flags);
      }, () -> {
         return this.spliterator(type, flags);
      });
   }

   default @NotNull Iterator<Component> iterator(final @NotNull ComponentIteratorType type, final @Nullable ComponentIteratorFlag... flags) {
      return this.iterator(type, flags == null ? Collections.emptySet() : MonkeyBars.enumSet(ComponentIteratorFlag.class, flags));
   }

   default @NotNull Iterator<Component> iterator(final @NotNull ComponentIteratorType type, final @NotNull Set<ComponentIteratorFlag> flags) {
      return new ComponentIterator(this, (ComponentIteratorType)Objects.requireNonNull(type, "type"), (Set)Objects.requireNonNull(flags, "flags"));
   }

   default @NotNull Spliterator<Component> spliterator(final @NotNull ComponentIteratorType type, final @Nullable ComponentIteratorFlag... flags) {
      return this.spliterator(type, flags == null ? Collections.emptySet() : MonkeyBars.enumSet(ComponentIteratorFlag.class, flags));
   }

   default @NotNull Spliterator<Component> spliterator(final @NotNull ComponentIteratorType type, final @NotNull Set<ComponentIteratorFlag> flags) {
      return Spliterators.spliteratorUnknownSize(this.iterator(type, flags), 1296);
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   @Contract(
      pure = true
   )
   default @NotNull Component replaceText(final @NotNull String search, final @Nullable ComponentLike replacement) {
      return this.replaceText((b) -> {
         b.matchLiteral(search).replacement(replacement);
      });
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   @Contract(
      pure = true
   )
   default @NotNull Component replaceText(final @NotNull Pattern pattern, final @NotNull Function<TextComponent.Builder, @Nullable ComponentLike> replacement) {
      return this.replaceText((b) -> {
         b.match(pattern).replacement(replacement);
      });
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   @Contract(
      pure = true
   )
   default @NotNull Component replaceFirstText(final @NotNull String search, final @Nullable ComponentLike replacement) {
      return this.replaceText((b) -> {
         b.matchLiteral(search).once().replacement(replacement);
      });
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   @Contract(
      pure = true
   )
   default @NotNull Component replaceFirstText(final @NotNull Pattern pattern, final @NotNull Function<TextComponent.Builder, @Nullable ComponentLike> replacement) {
      return this.replaceText((b) -> {
         b.match(pattern).once().replacement(replacement);
      });
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   @Contract(
      pure = true
   )
   default @NotNull Component replaceText(final @NotNull String search, final @Nullable ComponentLike replacement, final int numberOfReplacements) {
      return this.replaceText((b) -> {
         b.matchLiteral(search).times(numberOfReplacements).replacement(replacement);
      });
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   @Contract(
      pure = true
   )
   default @NotNull Component replaceText(final @NotNull Pattern pattern, final @NotNull Function<TextComponent.Builder, @Nullable ComponentLike> replacement, final int numberOfReplacements) {
      return this.replaceText((b) -> {
         b.match(pattern).times(numberOfReplacements).replacement(replacement);
      });
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   @Contract(
      pure = true
   )
   default @NotNull Component replaceText(final @NotNull String search, final @Nullable ComponentLike replacement, final @NotNull IntFunction2<PatternReplacementResult> fn) {
      return this.replaceText((b) -> {
         b.matchLiteral(search).replacement(replacement).condition(fn);
      });
   }

   /** @deprecated */
   @Deprecated
   @ScheduledForRemoval(
      inVersion = "5.0.0"
   )
   @Contract(
      pure = true
   )
   default @NotNull Component replaceText(final @NotNull Pattern pattern, final @NotNull Function<TextComponent.Builder, @Nullable ComponentLike> replacement, final @NotNull IntFunction2<PatternReplacementResult> fn) {
      return this.replaceText((b) -> {
         b.match(pattern).replacement(replacement).condition(fn);
      });
   }

   default void componentBuilderApply(final @NotNull ComponentBuilder<?, ?> component) {
      component.append(this);
   }

   default @NotNull Component asComponent() {
      return this;
   }

   default @NotNull HoverEvent<Component> asHoverEvent(final @NotNull UnaryOperator<Component> op) {
      return HoverEvent.showText((Component)op.apply(this));
   }

   default @NotNull Stream<? extends ExaminableProperty> examinableProperties() {
      return Stream.of(ExaminableProperty.of("style", this.style()), ExaminableProperty.of("children", this.children()));
   }
}
